from unittest.mock import Mock, call
from typing import ContextManager

import pytest

from classic.operations import Operation, operation, Cancel


class Service:

    def __init__(self, operation_: Operation):
        self.operation_ = operation_

    def work_with_context_manager(self):
        with self.operation_:
            return True

    @operation
    def work_with_decorator(self):
        return True

    @operation
    def just_fail(self):
        raise AssertionError

    def _add_callbacks(self, before, after, on_cancel, on_finish):
        self.operation_.before_complete(before)
        self.operation_.after_complete(after)
        self.operation_.on_cancel(on_cancel)
        self.operation_.on_finish(on_finish)

    @operation
    def success_with_callbacks(self, *args):
        self._add_callbacks(*args)
        return True

    @operation
    def failure_with_callbacks(self, *args):
        self._add_callbacks(*args)
        raise AssertionError

    @operation
    def success_with_nested_operation(self, *args):
        return self.success_with_callbacks(*args)

    def try_to_attach_callback_without_start(self, method_name):
        method = getattr(self.operation_, method_name)
        method(print)

    @operation
    def return_operation_in_progress_state(self):
        return self.operation_.in_progress

    def cancel_no_suppress(self):
        with self.operation_:
            raise self.operation_.Cancel

    def cancel_with_suppress(self):
        with self.operation_:
            raise Cancel(suppress=True)


class CM:
    def __init__(self):
        self.is_entered = False
        self.is_exited = False

    def __enter__(self):
        self.is_entered = True

    def __exit__(self, *args):
        self.is_exited = True
        return False


@pytest.fixture
def op():
    cm: ContextManager = Mock()
    cm.__enter__ = Mock()
    cm.__exit__ = Mock()

    return Operation(
        context_managers=[cm],
        before_start=[Mock()],
        after_start=[Mock()],
        before_complete=[Mock()],
        after_complete=[Mock()],
        on_cancel=[Mock()],
        on_finish=[Mock()],
    )


@pytest.fixture
def service(op):
    return Service(operation_=op)


@pytest.fixture
def callbacks():
    return [Mock(), Mock(), Mock(), Mock()]


def check_successful_call(operation_: Operation, times=1):
    calls = [call() for __ in range(times)]

    operation_._before_start[0].assert_has_calls(calls)

    cm = operation_._context_managers[0]
    cm.__enter__.assert_has_calls([
        call(cm) for __ in range(times)
    ])

    operation_._after_start[0].assert_has_calls(calls)

    operation_._before_complete[0].assert_has_calls(calls)

    cm.__exit__.assert_has_calls([
        call(cm, None, None, None) for __ in range(times)
    ])

    operation_._after_complete[0].assert_has_calls(calls)

    operation_._on_cancel[0].assert_not_called()

    operation_._on_finish[0].assert_has_calls(calls)


def check_failure_call(operation_: Operation, times=1):
    calls = [call() for __ in range(times)]

    operation_._before_start[0].assert_has_calls(calls)

    cm = operation_._context_managers[0]
    cm.__enter__.assert_has_calls([
        call(cm) for __ in range(times)
    ])

    operation_._after_start[0].assert_has_calls(calls)

    operation_._before_complete[0].assert_not_called()

    assert len(cm.__exit__.mock_calls) == times

    operation_._after_complete[0].assert_not_called()

    operation_._on_cancel[0].assert_has_calls(calls)

    operation_._on_finish[0].assert_has_calls(calls)


def test_operation_as_cm(service):
    result = service.work_with_context_manager()

    assert result is True
    check_successful_call(service.operation_)


def test_operation_as_decorator(service):
    result = service.work_with_decorator()

    assert result is True
    check_successful_call(service.operation_)


def test_operation_calls_callbacks_on_failure(service):
    with pytest.raises(AssertionError):
        service.just_fail()

    check_failure_call(service.operation_)


def test_dynamic_callback_on_success(service, callbacks):
    before, after, cancel, finish = callbacks

    result = service.success_with_callbacks(*callbacks)

    assert result is True
    check_successful_call(service.operation_)
    before.assert_called_once()
    after.assert_called_once()
    cancel.assert_not_called()
    finish.assert_called_once()


def test_dynamic_callback_on_failure(service, callbacks):
    before, after, cancel, finish = callbacks

    with pytest.raises(AssertionError):
        service.failure_with_callbacks(before, after, cancel, finish)

    check_failure_call(service.operation_)
    before.assert_not_called()
    after.assert_not_called()
    cancel.assert_called_once()
    finish.assert_called_once()


def test_nested_operation(service, callbacks):
    before, after, cancel, finish = callbacks

    result = service.success_with_nested_operation(
        before, after, cancel, finish,
    )

    assert result is True
    check_successful_call(service.operation_)
    before.assert_called_once()
    after.assert_called_once()
    cancel.assert_not_called()
    finish.assert_called_once()


def test_operation_is_reusable(service, callbacks):
    before, after, cancel, finish = callbacks

    result_1 = service.success_with_callbacks(before, after, cancel, finish)
    result_2 = service.success_with_callbacks(before, after, cancel, finish)

    assert result_1 is True
    assert result_2 is True
    check_successful_call(service.operation_, 2)
    before.assert_has_calls([call(), call()])
    after.assert_has_calls([call(), call()])
    cancel.assert_not_called()
    finish.assert_has_calls([call(), call()])


@pytest.mark.parametrize('method_name', [
    'before_complete',
    'after_complete',
    'on_cancel',
    'on_finish',
])
def test_dynamic_callbacks_works_only_after_start(service, method_name):
    with pytest.raises(AssertionError):
        service.try_to_attach_callback_without_start(method_name)


def test_in_progress_property(service):
    assert service.operation_.in_progress is False
    assert service.return_operation_in_progress_state() is True
    assert service.operation_.in_progress is False


def test_error_in_cm_enter(service, callbacks):
    before, after, cancel, finish = callbacks

    broken_cm = Mock()
    broken_cm.__enter__ = Mock(side_effect=AssertionError)
    broken_cm.__exit__ = Mock()

    service.operation_._context_managers = [broken_cm]

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_called_once()
    op._after_start[0].assert_not_called()
    op._before_complete[0].assert_not_called()
    op._context_managers[0].__exit__.assert_not_called()
    op._after_complete[0].assert_not_called()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_not_called()
    after.assert_not_called()
    cancel.assert_not_called()
    finish.assert_not_called()


def test_error_in_before_start(service, callbacks):
    before, after, cancel, finish = callbacks

    service.operation_._before_start[0] = Mock(side_effect=AssertionError)

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_not_called()
    op._after_start[0].assert_not_called()
    op._before_complete[0].assert_not_called()
    op._context_managers[0].__exit__.assert_not_called()
    op._after_complete[0].assert_not_called()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_not_called()
    after.assert_not_called()
    cancel.assert_not_called()
    finish.assert_not_called()


def test_error_in_after_start(service, callbacks):
    before, after, cancel, finish = callbacks

    service.operation_._after_start[0] = Mock(side_effect=AssertionError)

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_called_once()
    op._after_start[0].assert_called_once()
    op._before_complete[0].assert_not_called()
    op._context_managers[0].__exit__.assert_called_once()
    op._after_complete[0].assert_not_called()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_not_called()
    after.assert_not_called()
    cancel.assert_not_called()
    finish.assert_not_called()


def test_error_in_before_complete(service, callbacks):
    before, after, cancel, finish = callbacks

    service.operation_._before_complete[0] = Mock(side_effect=AssertionError)

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_called_once()
    op._after_start[0].assert_called_once()
    op._before_complete[0].assert_called_once()
    op._context_managers[0].__exit__.assert_called_once()
    op._after_complete[0].assert_not_called()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_not_called()
    after.assert_not_called()
    cancel.assert_called_once()
    finish.assert_called_once()


def test_error_in_cm_exit(service, callbacks):
    before, after, cancel, finish = callbacks

    broken_cm = Mock()
    broken_cm.__enter__ = Mock()
    broken_cm.__exit__ = Mock(side_effect=AssertionError)

    service.operation_._context_managers = [broken_cm]

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_called_once()
    op._after_start[0].assert_called_once()
    op._before_complete[0].assert_called_once()
    op._context_managers[0].__exit__.assert_called_once()
    op._after_complete[0].assert_not_called()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_called_once()
    after.assert_not_called()
    cancel.assert_called_once()
    finish.assert_called_once()


def test_error_in_after_complete(service, callbacks):
    before, after, cancel, finish = callbacks

    service.operation_._after_complete[0] = Mock(side_effect=AssertionError)

    with pytest.raises(AssertionError):
        service.success_with_callbacks(*callbacks)

    op = service.operation_
    op._before_start[0].assert_called_once()
    op._context_managers[0].__enter__.assert_called_once()
    op._after_start[0].assert_called_once()
    op._before_complete[0].assert_called_once()
    op._context_managers[0].__exit__.assert_called_once()
    op._after_complete[0].assert_called_once()
    op._on_cancel[0].assert_called_once()
    op._on_finish[0].assert_called_once()

    before.assert_called_once()
    after.assert_not_called()
    cancel.assert_called_once()
    finish.assert_called_once()


def test_instancing_with_no_iterable():
    context_manager = Mock()
    before_start = Mock()
    after_start = Mock()
    before_complete = Mock()
    after_complete = Mock()

    operation_ = Operation(
        context_managers=context_manager,
        before_start=before_start,
        after_start=after_start,
        before_complete=before_complete,
        after_complete=[after_complete],
        on_cancel=None,
    )

    assert operation_._context_managers == [context_manager]
    assert operation_._before_start == [before_start]
    assert operation_._after_start == [after_start]
    assert operation_._before_complete == [before_complete]
    assert operation_._after_complete == [after_complete]
    assert operation_._on_cancel == []
    assert operation_._on_finish == []


def test_cancel_with_suppress_false(service):
    with pytest.raises(Cancel):
        service.cancel_no_suppress()


def test_cancel_with_suppress_true(service):
    service.cancel_with_suppress()



from contextlib import ExitStack
from types import TracebackType
from typing import (
    List, Optional, Type, Iterable, ContextManager, Union, ClassVar
)
import threading

from .callbacks import Callback, Callbacks


OptionalContextManagers = Union[ContextManager, List[ContextManager], None]
OptionalCallbacks = Union[Callback, List[Callback], None]


def to_list(obj):
    if obj is None:
        return []
    elif isinstance(obj, Iterable):
        return list(obj)
    else:
        return [obj]


class Cancel(Exception):

    def __init__(self, suppress: bool = False):
        self.suppress = suppress


class Operation(threading.local):
    """
    Класс для выделения границ операций в приложении.

    Представляет собой контекстный менеджер, которым следует оборачивать
    действия пользователей в приложении. На границах операции, то есть входе и
    выходе в блок with, происходит вызов коллбеков и работа с вложенными
    контекстными менеджерами. Подробно жизненный цикл описан в README и
    docstring-ах методов.

    Потокобезопасен, локален для текущего потока.
    """

    _context_managers: List[ContextManager]
    _before_start: List[Callback]
    _after_start: List[Callback]
    _before_complete: List[Callback]
    _after_complete: List[Callback]
    _on_cancel: List[Callback]
    _on_finish: List[Callback]
    _exit_stack: ExitStack
    _calls_count: int
    _current: Optional[Callbacks]

    Cancel: ClassVar[Type[Cancel]] = Cancel

    def __init__(
        self,
        context_managers: OptionalContextManagers = None,
        before_start: OptionalCallbacks = None,
        after_start: OptionalCallbacks = None,
        before_complete: OptionalCallbacks = None,
        after_complete: OptionalCallbacks = None,
        on_cancel: OptionalCallbacks = None,
        on_finish: OptionalCallbacks = None,
    ):
        self._context_managers = to_list(context_managers)
        self._before_start = to_list(before_start)
        self._after_start = to_list(after_start)
        self._before_complete = to_list(before_complete)
        self._after_complete = to_list(after_complete)
        self._on_cancel = to_list(on_cancel)
        self._on_finish = to_list(on_finish)
        self._exit_stack = ExitStack()
        self._calls_count = 0
        self._current = None

    def _new_callbacks(self):
        """Порождает новый контейнер с callback-ами."""
        return Callbacks(
            before_complete=self._before_complete.copy(),
            after_complete=self._after_complete.copy(),
            on_cancel=self._on_cancel.copy(),
            on_finish=self._on_finish.copy(),
        )

    def __enter__(self):
        """
        Запуск операции.

        При первом обращении запускает коллбеки before_start, затем
        выполняет вход во все указанные контекстные менеджеры, затем
        запускает коллбеки after_start.

        Если что-то пошло не так, вызовет on_cancel, затем on_finish, выбросит
        исключение наружу.
        """
        if self._calls_count == 0:
            self._current = self._new_callbacks()

            try:
                self._handle_for_first_error(self._before_start)

                self._exit_stack.__enter__()
                for context in self._context_managers:
                    self._exit_stack.enter_context(context)

                self._handle_for_first_error(self._after_start)
            except Exception as exc:
                new_exc = None
                try:
                    self._exit_stack.__exit__(type(exc), exc, exc.__traceback__)
                except Exception as exc_:
                    new_exc = exc_
                try:
                    self._cancel()
                finally:
                    self._finish()
                raise new_exc or exc

        self._calls_count += 1

    def __exit__(
        self, exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Окончание операции.

        Вызывает коллбеки before_complete, потом совершает выход из контекстных
        менеджеров, затем вызывает after_complete.

        Если что-то пошло не так, вызовет on_cancel, затем on_finish, выбросит
        исключение наружу.
        """
        self._calls_count -= 1

        if self._calls_count != 0:
            return False

        suppress = False

        if exc_type is None:
            try:
                self._handle_for_first_error(
                    self._current.before_complete
                )
            except Exception as exc:
                exc_type = type(exc)
                exc_val = exc
                exc_tb = exc.__traceback__

        try:
            self._exit_stack.__exit__(exc_type, exc_val, exc_tb)
        except Exception as exc:
            exc_type = type(exc)
            exc_val = exc

        try:
            if exc_type is None:
                self._handle_for_first_error(
                    self._current.after_complete
                )
            else:
                # Используется raise для того, чтобы в случае исключения в
                # callback отработал _cancel()
                raise exc_val
        except Exception:
            self._cancel()
            raise
        finally:
            self._finish()
            if isinstance(exc_val, Cancel):
                return exc_val.suppress

        return False

    @property
    def in_progress(self) -> bool:
        """Сигнализирует о состоянии операции.
        True - запущена, False - не запущена.
        """
        return self._calls_count > 0

    def before_complete(self, callback: Callback):
        """Добавляет callback для вызова перед успешным завершением операции.
        Callback может быть любым вызываемым объектом
        и не должен принимать аргументов.
        """
        assert self.in_progress
        self._current.before_complete.append(callback)

    def after_complete(self, callback: Callback):
        """Добавляет callback для вызова после успешного завершения операции.
        Callback может быть любым вызываемым объектом
        и не должен принимать аргументов.
        """
        assert self.in_progress
        self._current.after_complete.append(callback)

    def on_cancel(self, callback: Callback):
        """Добавляет callback для вызова при отмене операции.
        Callback может быть любым вызываемым объектом
        и не должен принимать аргументов.
        """
        assert self.in_progress
        self._current.on_cancel.append(callback)

    def on_finish(self, callback: Callback):
        """Добавляет callback для вызова после окончания операции.
        Callback может быть любым вызываемым объектом
        и не должен принимать аргументов.
        """
        assert self.in_progress
        self._current.on_finish.append(callback)

    @staticmethod
    def _try_handle_all(callbacks: Iterable[Callback]):
        """Вызывает указанные callbacks последовательно,
        отлавливая и аккумулируя ошибки из callbacks.
        В случае, если ошибки все же произошли, выкинет RuntimeError,
        содержащий список ошибок.
        """
        errors = []
        for callback in callbacks:
            try:
                callback()
            except Exception as exc:
                errors.append(exc)

        if errors:
            raise RuntimeError(errors)

    @staticmethod
    def _handle_for_first_error(callbacks: Iterable[Callback]):
        """Попытка вызывать указанные callbacks последовательно,
        без отлова ошибок в callbacks.
        """
        for callback in callbacks:
            callback()

    def _cancel(self):
        """Отмена операции.
        Вызывает все callbacks, назначенные на отмену операции,
        независимо друг от друга.
        """
        self._try_handle_all(
            self._current.on_cancel
        )

    def _finish(self):
        """Завершение операции.
        Вызывает все callbacks, назначенные на завершение операции,
        чистит текущее состояние.
        """
        self._try_handle_all(
            self._current.on_finish
        )
        del self._current



from functools import wraps

from classic.components import add_extra_annotation, doublewrap

from .operation import Operation, Cancel


@doublewrap
def operation(method, prop: str = 'operation_', type_: Operation = Operation):
    """Сахар для облегчения применения операций.
    По сути просто оборачивает функцию в блок with c указанным полем из self.

    Если установлена библиотека classic.components, то добавляет операцию в
    дополнительные аннотации.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with getattr(self, prop):
            result = method(self, *args, **kwargs)

        return result

    return add_extra_annotation(wrapper, prop, type_)


operation.Cancel = Cancel



from typing import Callable, List


Callback = Callable[[], None]


class Callbacks:

    def __init__(
        self,
        before_complete: List[Callback] = None,
        after_complete: List[Callback] = None,
        on_cancel: List[Callback] = None,
        on_finish: List[Callback] = None,
    ):
        self.before_complete = before_complete or []
        self.after_complete = after_complete or []
        self.on_cancel = on_cancel or []
        self.on_finish = on_finish or []



from unittest.mock import Mock

from classic.operations import Operation, operation


class SomeService:

    def __init__(self, read: Operation):
        self.read = read

    @operation('read')
    def some_method(self):
        pass


def test_operation_customize():

    class CM:

        def __init__(self):
            self.mock = Mock()
            self.mock.__enter__ = Mock()
            self.mock.__exit__ = Mock()

        def __enter__(self):
            self.mock.__enter__(self)

        def __exit__(self, *args):
            self.mock.__exit__(self, *args)

    cm = CM()
    op = Operation(context_managers=cm)
    service = SomeService(read=op)

    service.some_method()

    cm.mock.__enter__.assert_called_once()
    cm.mock.__exit__.assert_called_once()
